using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Xml.Serialization;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace EQIA.EQIDL
{
    /// <summary>
    /// Information about a single frame of animation.
    /// </summary>
    [TypeConverter(typeof(FrameConverter))]
    public partial class Frame : Class, IReferer, ILoadable, IDrawable, IDrawableSurface
    {
        #region Members

        private TextureInfo textureInfo_Texture;
        private Texture2D mySurfaceTexture;
        private RenderTarget2D mySurfaceTarget;
        private bool isRequireRedraw;

        private string myTexture;
        private int myDuration;
        private Point myLocation, myHotspot;
        private Size mySize;
        private RGB myShading, mySpecular;

        #endregion

        #region Properties

        /// <summary>
        /// A texture in which this frame's image resides in.
        /// </summary>
        [Category("Appearance"), Description("Name of the texture this frame's image resides in.")]
        [XmlElement("Texture", IsNullable = false)]
        [NotifyParentProperty(true)]
        public string Texture { get { return myTexture; } set { if (myTexture != value) { isRequireRedraw = true; myTexture = value; } } }

        /// <summary>
        /// Amount of milliseconds this frame will be displayed before the animation progresses.
        /// </summary>
        [Category("Appearance"), Description("Amount of milliseconds this frame will be displayed before the animation progresses.")]
        [XmlElement("Duration", IsNullable = false)]
        [NotifyParentProperty(true)]
        public int Duration { get { return myDuration; } set { myDuration = value; } }

        /// <summary>
        /// Location of this frame's top-left corner in texture co-ordinates.
        /// </summary>
        [Category("Appearance"), Description("Location of this frame's top-left corner in texture co-ordinates.")]
        [XmlElement("Location", IsNullable = false)]
        [NotifyParentProperty(true)]
        public Point Location { get { return myLocation; } set { if (myLocation != value) { isRequireRedraw = true; myLocation = value; } } }

        /// <summary>
        /// A point used to keep the frame centered on the animation location.
        /// </summary>
        [Category("Appearance"), Description("A point used to keep the frame centered on the animation location.")]
        [XmlElement("Hotspot", IsNullable = false)]
        [NotifyParentProperty(true)]
        public Point Hotspot { get { return myHotspot; } set { myHotspot = value; } }

        /// <summary>
        /// Size in pixels of the image for this frame.
        /// </summary>
        [Category("Appearance"), Description("Size in pixels of the image for this frame.")]
        [XmlElement("Size", IsNullable = false)]
        [NotifyParentProperty(true)]
        public Size Size { get { return mySize; } set { if (mySize != value) { isRequireRedraw = true; mySize = value; } } }

        /// <summary>
        /// A color used to tint this frame's image when drawn.
        /// </summary>
        [Category("Appearance"), Description("A color used to tint this frame's image when drawn.")]
        [XmlElement(ElementName = "Shading", IsNullable = false)]
        [NotifyParentProperty(true)]
        public RGB Shading { get { return myShading; } set { if (value != myShading) { isRequireRedraw = true; myShading = value; } } }

        /// <summary>
        /// A gloss color used in specular highlights, similar to the white gleam of sunlight on a gloss coated car hood.
        /// </summary>
        [Category("Appearance"), Description("A gloss color used in specular highlights, similar to the white gleam of sunlight on a gloss coated car hood.")]
        [XmlElement(ElementName = "Specular", IsNullable = false)]
        [NotifyParentProperty(true)]
        public RGB Specular { get { return mySpecular; } set { if (value != mySpecular) { isRequireRedraw = true; mySpecular = value; } } }

        #endregion

        #region Creation

        public Frame()
            : this(null)
        {
        }

        public Frame(EQInterface eqInterface)
            : base(eqInterface)
        {
        }

        #endregion

        #region Deserialization

        protected override void SetProperties(object obj)
        {
            base.SetProperties(obj);

            Frame temp_Frame = obj as Frame;
            if (temp_Frame != null)
            {
                Texture = temp_Frame.Texture;
                Duration = temp_Frame.Duration;
                Location = temp_Frame.Location;
                Hotspot = temp_Frame.Hotspot;
                Size = temp_Frame.Size;
                Shading = temp_Frame.Shading;
                Specular = temp_Frame.Specular;
            }
        }

        #endregion

        #region IReferer Members

        public void ResolveReferences()
        {
            if (Interface != null && !string.IsNullOrEmpty(myTexture))
                textureInfo_Texture = Interface.GetElement<TextureInfo>(myTexture);
        }

        public bool ValidateReferences()
        {
            if (textureInfo_Texture != null && textureInfo_Texture.Name == myTexture)
                return true;
            return false;
        }

        #endregion

        #region ILoadable Members

        [XmlIgnore()]
#if DEBUG
        [Browsable(true), Category("Internal Data"), NotifyParentProperty(true)]
#else
        [Browsable(false)]
#endif
        public bool IsLoaded { get { return textureInfo_Texture != null && textureInfo_Texture.IsLoaded; } }

        public void LoadContent(GraphicsDevice device)
        {
            if (textureInfo_Texture != null && !textureInfo_Texture.IsLoaded)
                textureInfo_Texture.LoadContent(device);
        }

        public void UnloadContent()
        {
            if (textureInfo_Texture != null && textureInfo_Texture.IsLoaded) 
                textureInfo_Texture.UnloadContent();
        }

        #endregion

        #region IDrawable Members

        /// <summary>
        /// Draw this frame at Point.Zero at it's natural size.
        /// </summary>
        /// <param name="batch">Sprite batch that has already had Begin called.</param>
        public void Draw(SpriteBatch batch) { Draw(batch, Point.Zero, Size.Empty); }

        /// <summary>
        /// Draw this frame at the specified location with a given size.
        /// </summary>
        /// <param name="batch">Sprite batch that has already had Begin called.</param>
        /// <param name="location">Location to draw the top-left corner of this frame.</param>
        /// <param name="size">Size in pixels to draw the frame as. Use Size.Empty to draw in it's natural size.</param>
        public void Draw(SpriteBatch batch, Point location, Size size)
        {
            if (mySurfaceTexture !=  null && !mySurfaceTexture.IsDisposed)
            {
                if (size == Size.Empty) size = mySize;
                batch.Draw(mySurfaceTexture, new Rectangle(location.X, location.Y, size.Width, size.Height), Color.White);
            }
        }

        #endregion

        #region IDrawableSurface Members

        [XmlIgnore()]
#if DEBUG
        [Browsable(true), Category("Internal Data"), NotifyParentProperty(true)]
#else
        [Browsable(false)]
#endif
        public bool IsRequireRedraw { get { return isRequireRedraw; } }

        public void DrawToTexture(SpriteBatch batch)
        {
            if (mySurfaceTarget == null || mySurfaceTarget.IsDisposed)
                return;

            if (isRequireRedraw && textureInfo_Texture != null && textureInfo_Texture.IsLoaded)
            {
                RenderTarget2D pRenderTarget = (RenderTarget2D)batch.GraphicsDevice.GetRenderTarget(0);
                batch.GraphicsDevice.SetRenderTarget(0, mySurfaceTarget);
                batch.GraphicsDevice.Clear(Color.TransparentBlack);
                batch.Begin(SpriteBlendMode.AlphaBlend);

                textureInfo_Texture.Draw(batch, 
                    new Rectangle(0, 0, mySize.Width, mySize.Height),
                    new Rectangle(myLocation.X, myLocation.Y, mySize.Width, mySize.Height), 
                    RGB.White);//myShading);

                batch.End();
                batch.GraphicsDevice.SetRenderTarget(0, pRenderTarget);
                mySurfaceTexture = mySurfaceTarget.GetTexture();
                
                isRequireRedraw = false;
            }
        }

        public void EnsureRenderTarget(GraphicsDevice device)
        {
            if (mySize.Width < 1 || mySize.Height < 1) return;

            if (mySurfaceTarget == null || mySurfaceTarget.Width != mySize.Width || mySurfaceTarget.Height != mySize.Height)
            {
                try
                {
                    mySurfaceTarget = new RenderTarget2D(device, 
                        mySize.Width, mySize.Height, 1,
                        device.PresentationParameters.BackBufferFormat, MultiSampleType.None, 0);
                    isRequireRedraw = true;
                }
                catch { }
            }
        }

        #endregion
    }

    public class FrameConverter : ExpandableObjectConverter
    {
        public override object ConvertTo(ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value, Type destinationType)
        {
            return (destinationType == typeof(string)) ? string.Empty : base.ConvertTo(context, culture, value, destinationType);
        }
    }
}
